library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library altera_mf;
use altera_mf.all;

library stratix;
use stratix.all;


entity jddr_dqs_group is
  generic (
    gDLL_INPUT_FREQUENCY                : string  := "7500ps";
    gSTRATIX_DQS_PHASE                  : string  := "72";       --  "90" degrees,  "72" degrees
    gPOSTAMBLE_CYCLE                    : integer := 1         -- 0,1,2
  );
  port (
    -- globals
    clk                     : in   std_logic;
    reset_n                 : in   std_logic;
    dqs_outclk : in std_logic;
    write_clk               : in   std_logic;
    resynch_clk             : in   std_logic;
    postamble_clk           : in   std_logic;
    dqs_delay_ctrl          : in   std_logic;

    -- DDR Controller Signals
    doing_wr    : in   std_logic;
    wdata_low   : in   std_logic_vector(7 downto 0);
    wdata_high  : in   std_logic_vector(7 downto 0);
    doing_rd    : in   std_logic;
    rdata_low   : out  std_logic_vector(7 downto 0);
    rdata_high  : out  std_logic_vector(7 downto 0);
    -- DDR Interface Signals
    ddr_dq   : inout std_logic_vector(7 downto 0);
    ddr_dqs  : inout std_logic;
    ddr_dm   : out   std_logic
  );
end jddr_dqs_group;

architecture rtl of jddr_dqs_group is
  component stratix_lcell
    generic (
      operation_mode         : string := "normal";
      synch_mode             : string := "off";
      register_cascade_mode  : string := "off";
      sum_lutc_input         : string := "datac";
      lut_mask               : string := "ffff";
      power_up               : string := "low";
      cin0_used              : string := "false";
      cin1_used              : string := "false";
      cin_used               : string := "false";
      output_mode            : string := "comb_only";
      lpm_type               : string := "stratix_lcell"
    );
    port (
            -- synthesis translate_off
            devclrn    : in std_logic := '1';
            devpor     : in std_logic := '1';
            -- synthesis translate_on
            clk        : in std_logic := '0';
            dataa      : in std_logic := '1';
            datab      : in std_logic := '1';
            datac      : in std_logic := '1';
            datad      : in std_logic := '1';
            aclr       : in std_logic := '0';
            aload      : in std_logic := '0';
            sclr       : in std_logic := '0';
            sload      : in std_logic := '0';
            ena        : in std_logic := '1';
            cin        : in std_logic := '0';
            cin0       : in std_logic := '0';
            cin1       : in std_logic := '1';
            inverta    : in std_logic := '0';
            regcascin  : in std_logic := '0';
            combout    : out std_logic;
            regout     : out std_logic;
            cout       : out std_logic;
            cout0      : out std_logic;
            cout1      : out std_logic
        );
        end component;

    component  stratix_io
    generic
    (
        bus_hold                        :   string := "false";
        ddio_mode                       :   string := "none";
        extend_oe_disable               :   string := "false";
        input_async_reset               :    string := "none";
        input_power_up                  :    string := "low";
        input_register_mode             :    string := "none";
        input_sync_reset                :    string := "none";
        oe_async_reset                  :    string := "none";
        oe_power_up                     :    string := "low";
        oe_register_mode                :    string := "none";
        oe_sync_reset                   :    string := "none";
        open_drain_output               :    string := "false";
        operation_mode                  :    string;
        output_async_reset              :    string := "none";
        output_power_up                 :    string := "low";
        output_register_mode            :    string := "none";
        output_sync_reset               :    string := "none";
        sim_dll_phase_shift             :    string := "unused";
        sim_dqs_input_frequency         :    string := "unused";
        tie_off_oe_clock_enable         :    string := "false";
        tie_off_output_clock_enable     :    string := "false";
        lpm_type                        :    string := "stratix_io"
    );
    port
    (
        areset              :    in std_logic := '0';
        combout             :    out std_logic;
        datain              :    in std_logic := '0';
        ddiodatain          :    in std_logic := '0';
        ddioregout          :    out std_logic;
        delayctrlin         :    in std_logic := '0';
        dqsundelayedout     :    out std_logic;
        inclk               :    in std_logic := '0';
        inclkena            :    in std_logic := '1';
        oe                  :    in std_logic := '1';
        outclk              :    in std_logic := '0';
        outclkena           :    in std_logic := '1';
        padio               :    inout std_logic;
        regout              :    out std_logic;
        sreset              :    in std_logic := '0'
    );
    end component;


    -- data path signals
    signal dqs_clk              : std_logic;
    signal dqs_oe_r             : std_logic;
    signal dq_oe                : std_logic;

    signal dq_captured_rising   : std_logic_vector(7 downto 0);
    signal dq_captured_falling  : std_logic_vector(7 downto 0);

    signal dq_capture_clk       : std_logic; -- this is the clock that is used to capture the incoming read data, can be either DQS or a user supplied clock
    signal undelayed_dqs        : std_logic;
    signal dq_enable            : std_logic;

    signal wdata_r_low, wdata_r_high : std_logic_vector(7 downto 0);

    signal reset                : std_logic; -- inverse of reset_n

begin
-------------------------------------------------------------------------------
-- DM pins and their logic
--
-- We will only support full 72bit writes
-------------------------------------------------------------------------------
  ddr_dm <= '0';

-------------------------------------------------------------------------------
-- DQS pin and its logic
-------------------------------------------------------------------------------
    p_dqs_oe : process (reset_n, clk)
    begin
        if (reset_n = '0') then
            dqs_oe_r <= '0';
        elsif falling_edge(clk) then
            dqs_oe_r <= doing_wr;
        end if;
    end process p_dqs_oe;

    -- DQS pin
    dqs_io :  stratix_io
    GENERIC MAP (
        DDIO_MODE                     => "output",
        EXTEND_OE_DISABLE             => "true",
        OE_ASYNC_RESET                => "none",
        OE_POWER_UP                   => "low",
        OE_REGISTER_MODE              => "register",
        OE_SYNC_RESET                 => "none",
        OPEN_DRAIN_OUTPUT             => "false",
        OPERATION_MODE                => "bidir",
        OUTPUT_ASYNC_RESET            => "none",
        OUTPUT_POWER_UP               => "low",
        OUTPUT_REGISTER_MODE          => "register",
        OUTPUT_SYNC_RESET             => "none",
        SIM_DLL_PHASE_SHIFT           => gSTRATIX_DQS_PHASE,
        SIM_DQS_INPUT_FREQUENCY       => gDLL_INPUT_FREQUENCY,
        TIE_OFF_OE_CLOCK_ENABLE       => "false",
        TIE_OFF_OUTPUT_CLOCK_ENABLE   => "false"
    )
    PORT MAP (
        combout           => dqs_clk,
        datain            => dqs_oe_r,
        ddiodatain        => '0',
        delayctrlin       => dqs_delay_ctrl,
        dqsundelayedout   => undelayed_dqs,
        oe                => dqs_oe_r,
        outclk            => dqs_outclk,
        outclkena         => '1',
        padio             => ddr_dqs
    );

-------------------------------------------------------------------------------
-- DQ pins and their logic
-------------------------------------------------------------------------------
    dq_capture_clk <= not dqs_clk;

    p_dq_oe : process (reset_n, clk)
    begin
        if (reset_n = '0') then
            dq_oe <= '0';
        elsif rising_edge(clk) then
            dq_oe <= doing_wr;
        end if;
    end process p_dq_oe;

    g_dq_io : for i in 0 to 7 generate
    begin

        dq_io :  stratix_io
        generic map (
            ddio_mode               => "bidir",
            extend_oe_disable       => "false",
            input_async_reset       => "clear",
            input_power_up          => "low",
            input_register_mode     => "register",
            oe_async_reset          => "clear",
            oe_power_up             => "low",
            oe_register_mode        => "register",
            operation_mode          => "bidir",
            output_async_reset      => "clear",
            output_power_up         => "low",
            output_register_mode    => "register"
        )
        PORT MAP (
            inclkena    => dq_enable,
            areset      => reset,
            datain      => wdata_r_low(i),
            ddiodatain  => wdata_r_high(i),
            ddioregout  => dq_captured_rising(i),
            inclk       => dq_capture_clk,
            oe          => dq_oe,
            outclkena   => '1',
            outclk      => write_clk,
            padio       => ddr_dq(i),
            regout      => dq_captured_falling(i)
        );

    end generate; 

    reset <= not reset_n;

-------------------------------------------------------------------------------
-- Logic to disable the capture registers (particularly during DQS postamble)
--
-- The output of the dq_enable_reset register holds the dq_enable register in
-- reset (which *enables* the dq capture registers). The controller releases
-- the dq_enable register so that it is clocked by the last falling edge of the
-- read dqs signal. This disables the dq capture registers during and after the
-- dqs postamble so that the output of the dq capture registers can be safely
-- resynchronised.
--
-- Postamble Clock Cycle  : 1
-- Postamble Clock Edge   : rising
-- Postamble Regs Per DQS : 1
--
-------------------------------------------------------------------------------
  dq_enable <= '1';

--     dqs_buf1 : stratix_lcell
--         generic map (
--            operation_mode         => "normal",
--             lut_mask               => "ff00"
--         )
--         port map (
--             datad     => undelayed_dqs,
--             combout   => dqs_postamble_clk
--         );


    -- Critical registers clocked on the falling edge of the DQS to
    -- disable the DQ capture registers during the DQS postamble
--    p_dq_enable : process (dq_enable_reset, dqs_postamble_clk)
--    begin
--        if (dq_enable_reset = '1') then
--            dq_enable <= '1'; -- hold high when in reset so capture regs are *enabled*
--        elsif falling_edge(dqs_postamble_clk) then
--            dq_enable <=  '0'; -- disable capture regs on the last falling edge of dqs
--        end if;
--    end process;


    -- Use a rising edge for postamble
    -- The registers which generate the reset signal to the above registers
    -- Can be clocked off the resynch or system clock
--    g_rising_edge_postamble_clk0 : process (reset_n, postamble_clk)
--    begin
--        if (reset_n = '0') then
--            dq_enable_reset <= '0';
--        elsif rising_edge(postamble_clk) then
--            dq_enable_reset <= doing_rd_delayed;
--        end if;
--    end process;


    -- pipeline the doing_rd signal to enable and disable the DQ capture regs at the right time
--    process (reset_n, clk)
--    begin
--        if (reset_n = '0') then
--            doing_rd_pipe <= (others => '0');
--        elsif rising_edge(clk) then
--            doing_rd_pipe <= doing_rd_pipe(doing_rd_pipe'high-1 downto 0) & doing_rd; -- shift bits up
--        end if;
--    end process;


    -- It's safe to clock from falling edge of clk to postamble_clk, so use falling edge clock
--    process (reset_n, clk)
--   begin
--        if (reset_n = '0') then
--           doing_rd_delayed <= '0';
--        elsif falling_edge(clk) then
--            doing_rd_delayed <= doing_rd_pipe(gPOSTAMBLE_CYCLE);
--        end if;
--    end process;


-------------------------------------------------------------------------------
-- Write data registers
--
-- These are the last registers before the registers in the altddio_bidir. They
-- are clocked off the system clock but feed registers which are clocked off the
-- write clock, so their output is the beginning of 3/4 cycle path.
-------------------------------------------------------------------------------
    process (reset_n, clk)
    begin
        if (reset_n = '0') then
            wdata_r_low <= (others => '0');
            wdata_r_high <= (others => '0');
        elsif rising_edge(clk) then
            wdata_r_low <= wdata_low;
            wdata_r_high <= wdata_high;
        end if;
    end process;

    -- Use a rising edge for resynch
    p_rising_edge_resynch : process(reset_n, resynch_clk)
    begin
        if (reset_n = '0') then
            rdata_low <= (others => '0');
            rdata_high <= (others => '0');
        elsif rising_edge(resynch_clk) then
            rdata_low <= dq_captured_falling;
            rdata_high <= dq_captured_rising;
        end if;
    end process;

end rtl;

